/*
  This is an implementation of AWOBST-algorithm that constructs on approximately
  optimal binary search tree. That is, if we have n items {1,...,i,...,n} and
  each item has a weight wi, then the depth of an item i is less than
    log2 W - log2 wi + 1,
  where W is the sum of all weights W= SUM_{i=1}^n wi. This implementation is
  for the purpose of demonstrating the algorithm and is not very refined :-)

  The run time is O(n) in the unit cost models and also in low in IO-models, 
  because the algorithm performs linear scans and otherwise only a stack is 
  used to store temporary subtrees. Note that to achieve theoretical 
  guarantees, you need to override the dynamic memory allocation which I have 
  not done, because it is not necessarily useful.

  Note that:
  1) W must fit inside unsigned type
  2) The memory layout of the output is not optimized for IO-efficiency

  Copyright by Jussi Kujala. 
  Licenced under the GPL. For more information see http://www.gnu.org.
*/

#include <cassert>
#include <limits.h>
#include <stack>
#include "wbst.h"

//contains a tree and the depth restriction
typedef struct{Node* bst; int depth;} Instack; 

using namespace std;

//software implementation of the first set bit in argument
//returns the position of the _most_ significant bit, with position 0 being
//the _least_ significant bit
static inline unsigned FIRST_SET_BITsw(unsigned x)
{
  //32-bit implementation only
  assert(sizeof(unsigned)==4);

  unsigned r(0);
  //does a binary search over the bits
  for(int i=4;i>=0;i--)
  {
    unsigned t1= 1<<i;
    unsigned t2= 1<<t1;
    //true if there is a set bit in the upper part of x
    if(t2<=x)
    {
      r+=t1;
      x=(x>>t1);
    }
  }
  return r;
}

//x86 implementation of the first set bit in argument
static inline unsigned FIRST_SET_BITx86hw(unsigned x)
{
  assert(x!=0); //instruction bsf "fails" if x is 0
  asm("bsrl %0,%0"
      :"=r"(x)
      :"r"(x)
      :
     );
  return x;
}

static inline int FIRST_SET_BIT(int x)
{  
  //return FIRST_SET_BITx86hw(x); //Use the hw-version if possible
  return FIRST_SET_BITsw(x);  //When hw-version is not implemented
}

//rescales item weights so that they sum to something close to 2^(unsigned length)
unsigned rescale(vector<Item>& items, int W)
{
  unsigned target= 1<<(8*sizeof(unsigned)-1);
  double rs=((double)target)/((double)W); //the scaling factor
  W=0; //reset the sum of weights
  assert(rs>=1.0);//otherwise might get zero weights...
  vector<Item>::iterator ite=items.begin();
  while(ite!=items.end()) //linear scan over items
  {
    (*ite).weight=(unsigned) rs*(*ite).weight; //scales the weight of this item
    W+=(*ite).weight;
    ite++;
  }
  return W; //returns the new sum of weights
}

unsigned compute_weight_sum(vector<Item>& a_items) 
{
  unsigned W=0;
  vector<Item>::iterator ite=a_items.begin();
  while(ite!=a_items.end())
  {
    assert((*ite).weight>0); //positive weights only
    W+=(*ite).weight;
    ite++;
  }
  return W;
}

//The main algorithm that constructs the tree when given the items as input
//Assumes that the input is sorted according to the key values
Node* wbst(vector<Item>& a_items)
{
  assert(a_items.size()>0);
  
  //compute W the sum of weights
  unsigned W=compute_weight_sum(a_items);
  
  W=rescale(a_items,W); //scale item weights

  a_items.back().weight--; //decrease the weight of the last item slightly
  assert(a_items.back().weight > 0);

  //the main algorithm follows...
    
  stack<Instack> bst_stack; //contains (bst,depth)
  unsigned S(0); //maintains a sum of weights of items seen so far
  int prev_key=INT_MIN;
  vector<Item>::iterator ite=a_items.begin();
  while(ite!=a_items.end()) //linear scan over items
  {
    assert((*ite).key > prev_key);//check that input is sorted
    prev_key=(*ite).key;

    //obtain the depth restriction for this item
    int d=-FIRST_SET_BIT(S ^ (S+(*ite).weight) );
    
    //link trees in the stack if necessary
    while(bst_stack.size()>=2)
    {
      //pop two topmost trees
      Instack top1=bst_stack.top();
      bst_stack.pop();
      Instack top2=bst_stack.top();
      bst_stack.pop();
      //if it is possible to link them according to information we know 
      //then link them...
      if(top2.depth < top1.depth && d<top2.depth)
      {
        top2.bst->r=top1.bst;
        bst_stack.push(top2);
      }
      else //...otherwise put them back :-)
      {
        bst_stack.push(top2);
        bst_stack.push(top1);
        break;
      }
    }
    
    //note that the following line defines how items are located in the memory
    //allocation can be deferred until the parent of the node is known if you
    //want to do fewer I/O-operations (and even then you need to custom new)
    Node* current=new Node((*ite).key,0,0); 
    //Link the topmost tree in the stack as a left child if it is marked to 
    //belong to lower part of the tree
    if(!bst_stack.empty() && d < bst_stack.top().depth)
    {
      Instack top=bst_stack.top();
      bst_stack.pop();
      current->l=top.bst;
    }

    //Push the current tree to the stack
    Instack pushme;
    pushme.bst=current;
    pushme.depth=d;
    bst_stack.push(pushme);

    S+=(*ite).weight;
    ite++;
  }
  //link trees that are still in the stack
  while(bst_stack.size()>=2)
  {
    Instack top=bst_stack.top();
    bst_stack.pop();
    bst_stack.top().bst->r=top.bst;
  }
  Node* root=bst_stack.top().bst;
  bst_stack.pop();
  return root;
}

